import {createWorker, PSM, ImageLike, Page, Word, Bbox, Line, Paragraph} from 'tesseract.js';
import throttle from 'lodash.throttle';

type ProgressListener = (event: { status: string }) => void;
type AppendHtmlOptions = {
    adjustFontSize?: boolean;
    scale?: number;
}
type Options = {
    language?: string,
    lineClassName?: string,
    wordClassName?: string
}

export default class Ocr {
    data: Page = null;
    language: string = null;
    lineClassName: string = '';
    wordClassName: string = '';
    progressListeners: ProgressListener[] = []
    minFontSize = 2;

    constructor({language = 'eng', lineClassName = '', wordClassName = ''}: Options = {}) {
        this.language = language;
        this.lineClassName = lineClassName;
        this.wordClassName = wordClassName;
    }

    addSelectionListener() {
        document.addEventListener('selectionchange', throttle(this.handleSelection, 100));
        document.addEventListener('mousedown', this.handleMouseDown);
        document.addEventListener('copy', this.handleCopy);
    }

    removeSelectionListener() {
        document.removeEventListener('selectionchange', throttle(this.handleSelection, 100));
        document.removeEventListener('mousedown', this.handleMouseDown);
        document.removeEventListener('copy', this.handleCopy);
    }

    addProgressListener(listener: ProgressListener) {
        this.progressListeners.push(listener);
    }

    removeProgressListener(listener: ProgressListener) {
        this.progressListeners = this.progressListeners.filter(item => item !== listener);
    }

    /**
     * Parse image with tesseract
     *
     * @param {ImageLike} image The image instance.
     */
    async parseImage(image: ImageLike) {
        const worker = createWorker({
            logger: m => {
                this.progressListeners.forEach(callback => {
                    callback(m)
                })
            },
        });
        await worker.load();
        await worker.loadLanguage(this.language);
        await worker.initialize(this.language);
        await worker.setParameters({
            tessedit_pageseg_mode: PSM.AUTO_ONLY,
        });
        const {data} = await worker.recognize(image);
        this.data = data;
        await worker.terminate();
    }

    private updateTextSize(el: HTMLElement) {
        const wrap = document.createElement('span');
        document.body.appendChild(wrap);

        wrap.innerHTML = el.textContent.trim();
        const height = parseInt(el.style.height.replace('px', ''));
        const width = parseInt(el.style.width.replace('px', ''));
        let fontSize = height;
        if (fontSize > this.minFontSize) {
            wrap.style.fontSize = `${fontSize}px`;
            const fontSizeH = fontSize * height / wrap.offsetHeight;
            const fontSizeW = fontSize * width / wrap.offsetWidth;
            fontSize = fontSizeH < fontSizeW ? fontSizeH : fontSizeW;
        }
        if (fontSize < this.minFontSize) {
            fontSize = this.minFontSize;
        }
        el.style.fontSize = `${fontSize}px`;

        wrap.remove();
    }

    private updatePosition(el: HTMLElement, bbox: Bbox, scale: number, parentBbox?: Bbox) {
        const {x0, x1, y0, y1} = bbox;
        el.style.position = 'absolute';
        el.style.left = `${parentBbox ? (parentBbox.x0 * scale) : (x0 * scale)}px`;
        el.style.top = `${parentBbox ? 0 : (y0 * scale)}px`;
        el.style.width = `${(x1 - x0) * scale}px`;
        el.style.height = `${(y1 - y0) * scale}px`;
    }

    private updateElementSize(el: HTMLElement, {x0, x1, y0, y1}: Bbox, scale: number,) {
        el.style.width = `${(x1 - x0) * scale}px`;
        el.style.height = `${(y1 - y0) * scale}px`;
    }

    private mapWords(words: Word[], scale: number, adjustFontSize: boolean, parentBbox: Bbox) {
        const wordElementsList = [];

        for (let i = 0; i < words.length; i++) {
            const {bbox, text} = words[i];
            const nextBboxX = i < words.length - 1 ? words[i + 1].bbox.x0 : bbox.x1;
            const elementBbox = {x0: bbox.x0, x1: nextBboxX, y0: parentBbox.y0, y1: parentBbox.y1}

            if (!text.trim().length) continue;

            const el = document.createElement('span');
            this.updateElementSize(el, elementBbox, scale);
            el.innerText = text;
            if (this.wordClassName.length) {
                el.classList.add(this.wordClassName);
            }

            if (adjustFontSize) {
                this.updateTextSize(el)
            }

            wordElementsList.push(el);
        }

        return wordElementsList;
    }

    private mapLine(line: Line, scale: number, adjustFontSize: boolean) {
        const {bbox, words} = line;

        const el = document.createElement('span');
        this.updateElementSize(el, bbox, scale);
        el.style.display = 'flex';

        if (this.lineClassName.length) {
            el.classList.add(this.lineClassName);
        }

        const wordElements = this.mapWords(words, scale, adjustFontSize, bbox);

        if (!wordElements.length) return null;

        wordElements.forEach(word => {
            el.appendChild(word);
        })

        return {element: el, bbox};
    }

    mapParagraph(paragraph: Paragraph, scale: number) {
        const {bbox, lines} = paragraph;

        const el = document.createElement('p');
        this.updatePosition(el, bbox, scale);
        el.style.margin = '0px';

        const lineElements: { element: HTMLSpanElement, bbox: Bbox }[] = [];
        for (let i = 0; i < lines.length; i++) {
            const line = this.mapLine(lines[i], scale, true);
            if (line) {
                lineElements.push(line);
            }
        }

        if (!lineElements.length) return null;

        for (let i = 0; i < lineElements.length; i++) {
            const {element, bbox} = lineElements[i];
            if (i > 0) {
                const marginTop = bbox.y0 - lineElements[i - 1].bbox.y1;
                element.style.marginTop = `${marginTop}px`;
            }
            el.appendChild(element);
        }

        return {element: el, bbox};
    }

    /**
     * Append generated html to the container.
     *
     * @param {HTMLElement} container The container where elements should append.
     * @param {AppendHtmlOptions} options The options for customization.
     */
    appendHtml(container: HTMLElement, {adjustFontSize = true, scale = 1}: AppendHtmlOptions = {}) {
        if (!this.data.hocr) return;

        let elements: { bbox: Bbox, element: HTMLParagraphElement }[] = []
        const {paragraphs} = this.data;
        for (let i = 0; i < paragraphs.length; i++) {
            const paragraph = this.mapParagraph(paragraphs[i], scale)
            if (paragraph) {
                elements.push(paragraph);
            }
        }

        for (let i = 0; i < elements.length; i++) {
            const {element, bbox} = elements[i];

            if (i < elements.length - 1) {
                const padding = elements[i + 1].bbox.y0 - bbox.y1;
                if (padding > 0) {
                    element.style.paddingBottom = `${padding}px`;
                }
            }

            container.appendChild(element)
        }
    }

    handleSelection() {
        const selection = document.getSelection();
        if (selection?.focusNode?.parentNode) {
            const selectableElements = document.querySelectorAll('.selectable-element');
            for (const element of Array.from(selectableElements)) {
                if (!!selection.toString() && selection.containsNode(element, true)) {
                    (element as HTMLElement).style.background = '#ACCEF7';
                } else {
                    (element as HTMLElement).style.background = 'none';
                }
            }
        }
    }

    handleMouseDown() {
        const selection = document.getSelection();
        if (!selection) return;
        const selectableElements = document.querySelectorAll('.selectable-element');
        for (const element of Array.from(selectableElements)) {
            if (selection.containsNode(element, true)) {
                (element as HTMLElement).style.background = 'none';
            }
        }
    }

    handleCopy(event) {
        const selection = document.getSelection();
        if (!selection) return;
        const selectableElements = document.querySelectorAll('.selectable-element');
        let output = '';
        for (const element of Array.from(selectableElements)) {
            if (selection.containsNode(element, true)) {
                output+=`${(element as HTMLElement).textContent} `;
            }
        }
        event.clipboardData.setData('text/plain', output);
        event.preventDefault();
    }
}
